"""Classifier Resource Module."""

from ..dataclasses.layout_element_classifier import LayoutElementClassifierMeta
from .base import AsyncResource, SyncResource

from typing import Dict

class SyncMultiModalClassifierResource(SyncResource):
    """Synchronous Multimodal embedding and classification resource."""

    def classify(
        self,
        page_properties: Dict[int, dict],
    ) -> LayoutElementClassifierMeta:
        
        """Multimodal model classification of page layout"""
        task = "layout-element-classification"
        input_data = {"page_properties": page_properties}
            
        output = self._post(
            data={
                "input_data": input_data,
                "task": task,
            },
        )
        
        output.raise_for_status()
        return LayoutElementClassifierMeta(
            classified_blocks_by_page=output.json()["output"]
            )


class AsyncMultiModalClassifierResource(AsyncResource):
    """Asynchronous Multimodal embedding and classification resource"""

    async def classify(
        self,
        page_properties: Dict[int, dict],
        read_timeout: float = 10.0,
        timeout: float = 180.0,
    ) -> LayoutElementClassifierMeta:
        """Embed all texts."""
        task = "layout-element-classification"
        input_data = {"page_properties": page_properties}
            
        output = await self._post(
            data={
                "input_data": input_data,
                "task": task,
            },
            read_timeout=read_timeout,
            timeout=timeout
        )
        
        output.raise_for_status()
        return LayoutElementClassifierMeta(
            classified_blocks_by_page=output.json()["output"]
            )



"""Dataclasses for ranker models."""

from typing import Tuple

from pydantic import BaseModel


class RankerMeta(BaseModel):
    """Metadata for a ranker output."""

    pair: Tuple[str, str]
    score: float



"""Dataclasses for Quasar."""

from typing import Dict, Optional

from pydantic import BaseModel


class ModelData(BaseModel):
    """Model data from Quasar."""

    id: str
    provider: str
    metadata: Optional[Dict]



"""Extract Resource Module."""

from typing import Literal, Optional, Union

from ..dataclasses.extract import ExtractMeta, Keyword
from .base import AsyncResource, SyncResource


class SyncExtractorResource(SyncResource):
    """Synchronous ExtractorResource Class."""

    def extract(
        self,
        text: str,
        task: Optional[Union[Literal["topics"], Literal["keyword-extraction"]]] = None,
        model: Optional[str] = None,
        priority: int = 0,
        **kwargs,
    ) -> ExtractMeta:
        """Extract keywords or topics."""
        data = {
            "input_data": {"docs": [text], **kwargs},
            "priority": priority,
        }
        if model:
            data["model"] = model
        elif task:
            data["task"] = task
        else:
            raise ValueError("Either `task` or `model` must be provided.")
        output = self._post(data=data)
        output.raise_for_status()
        extract_response = output.json()["output"]
        return ExtractMeta(
            text=text,
            keywords=[
                Keyword(
                    keyword=kw["keyword"],
                    score=kw["score"],
                )
                for kw in extract_response[0]["keywords"]
            ],
        )


class AsyncExtractorResource(AsyncResource):
    """Asynchronous Extractor Resource Class."""

    async def extract(
        self,
        text: str,
        task: Optional[Union[Literal["topics"], Literal["keyword-extraction"]]] = None,
        model: Optional[str] = None,
        read_timeout: float = 10.0,
        timeout: float = 180.0,
        priority: int = 0,
        **kwargs,
    ) -> ExtractMeta:
        """Extract keywords or topics."""
        data = {
            "input_data": {"docs": [text], **kwargs},
            "task": task,
            "priority": priority,
        }
        if model:
            data["model"] = model
        elif task:
            data["task"] = task
        else:
            raise ValueError("Either `task` or `model` must be provided.")
        output = await self._post(
            data=data,
            read_timeout=read_timeout,
            timeout=timeout,
        )
        output.raise_for_status()
        extract_response = output.json()["output"]
        return ExtractMeta(
            text=text,
            keywords=[
                Keyword(
                    keyword=kw["keyword"],
                    score=kw["score"],
                )
                for kw in extract_response[0]["keywords"]
            ],
        )



"""Dataclasses for embeddings."""
from typing import Dict, List
from pydantic import BaseModel


class LayoutElementClassifierMeta(BaseModel):
    """Metadata for a multi modal embedding output."""

    classified_blocks_by_page: Dict[int, List[dict]]


"""Classifier Resource Module."""

from typing import List, Tuple

from ..dataclasses.column_count_classifier import ColumnCountClassifierMeta
from .base import AsyncResource, SyncResource


class SyncPageColumnCounterResource(SyncResource):
    """
    Synchronous Page Column Counter Resource Class.
    pages_images_base64_string: List of base64 encoded images of page, 
    bboxes: List of bounding boxes of words in the page (order: x0, y0, x1, y1),
    words: List of words in the page
    """

    def classify(
        self,
        pages_images_base64_string: List[str],
        bboxes: List[List[Tuple[int, int, int, int]]], 
        words: List[List[str]],
    ) -> ColumnCountClassifierMeta:
        """Column classification of every page"""
        
        task = "column-count-classification"
        input_data = {
                "pages_images_base64_string": pages_images_base64_string, 
                "bboxes": bboxes, 
                "words": words}
        
        output = self._post(
            data={
                "input_data": input_data,
                "task": task,
            },
        )
        
        output.raise_for_status()
        return ColumnCountClassifierMeta(
            column_counts=output.json()["output"]
        )


class AsyncPageColumnCounterResource(AsyncResource):
    """
    Asynchronous Page Column Counter Resource Class.
    pages_images_base64_string: List of base64 encoded images of page, 
    bboxes: List of bounding boxes of words in the page (order: x0, y0, x1, y1),
    words: List of words in the page
    """

    async def classify(
        self,
        pages_images_base64_string: List[str],
        bboxes: List[List[Tuple[int, int, int, int]]], 
        words: List[List[str]],
        read_timeout: float = 10.0,
        timeout: float = 180.0,
    ) -> ColumnCountClassifierMeta:
        """Column classification of every page"""
        
        task = "column-count-classification"
        input_data = {
            "pages_images_base64_string": pages_images_base64_string, 
            "bboxes": bboxes, 
            "words": words
        }
            
        output = await self._post(
            data={
                "input_data": input_data,
                "task": task,
            },
            read_timeout=read_timeout,
            timeout=timeout
        )
        
        output.raise_for_status()
        return ColumnCountClassifierMeta(
            column_counts=output.json()["output"]
        )



"""Classifier Resource Module."""

from typing import List, Literal, Optional, Union

from ..dataclasses.classify import Classification, ClassifierMeta
from .base import AsyncResource, SyncResource


class SyncClassifierResource(SyncResource):
    """Synchronous Classififer Resource Class."""

    def classify(
        self,
        text: str,
        task: Union[Literal["zero-shot-classification"], Literal["text-classification"]],
        labels: Optional[List[str]] = None,
        priority: int = 0,
    ) -> ClassifierMeta:
        """Embed all texts."""
        input_data = {"text": text}
        if task == "zero-shot-classification":
            input_data["candidate_labels"] = labels
        output = self._post(
            data={
                "input_data": input_data,
                "task": task,
                "priority": priority,
            },
        )
        output.raise_for_status()
        classification_resp = output.json()["output"]
        labels = classification_resp["labels"]
        scores = classification_resp["scores"]
        return ClassifierMeta(
            classifications=[
                Classification(label=label, confidence=score)
                for label, score in zip(labels, scores)
            ],
            text=text,
        )


class AsyncClassifierResource(AsyncResource):
    """Asynchronous Classifier Resource Class."""

    async def classify(
        self,
        text: str,
        task: Union[Literal["zero-shot-classification"], Literal["text-classification"]],
        labels: Optional[List[str]] = None,
        read_timeout: float = 10.0,
        timeout: float = 180.0,
        priority: int = 0,
    ) -> ClassifierMeta:
        """Embed all texts."""
        input_data = {"text": text}
        if task == "zero-shot-classification":
            input_data["candidate_labels"] = labels
        output = await self._post(
            data={
                "input_data": input_data,
                "task": task,
                "priority": priority,
            },
            read_timeout=read_timeout,
            timeout=timeout,
        )
        output.raise_for_status()
        classification_resp = output.json()["output"]
        labels = classification_resp["labels"]
        scores = classification_resp["scores"]
        return ClassifierMeta(
            classifications=[
                Classification(label=label, confidence=score)
                for label, score in zip(labels, scores)
            ],
            text=text,
        )



"""Dataclasses for embeddings."""
from typing import List
from pydantic import BaseModel


class ColumnCountClassifierMeta(BaseModel):
    """Metadata for a multi modal embedding output."""

    column_counts: List[int]


"""Tagger Resource Module."""

from typing import List, Literal, Optional, Union

from ..dataclasses.tag import TaggerMeta
from .base import AsyncResource, SyncResource


class SyncTaggerResource(SyncResource):
    """Synchronous Tagger Resource Class."""

    def tag(
        self,
        text: str,
        task: Optional[Union[Literal["ner"], Literal["acronym-detection"]]] = None,
        model: Optional[str] = None,
        priority: int = 0,
    ) -> List[TaggerMeta]:
        """Tag a text."""
        data = {
            "input_data": {"docs": [text]},
            "priority": priority,
        }
        if model:
            data["model"] = model
        elif task:
            data["task"] = task
        else:
            raise ValueError("Either `task` or `model` must be provided.")
        output = self._post(data=data)
        output.raise_for_status()
        return [
            TaggerMeta(
                end=entity["char_end_index"],
                label=entity["label"],
                score=entity["score"],
                span=entity["span"],
                start=entity["char_start_index"],
            )
            for entity in output.json()["output"][0]
        ]


class AsyncTaggerResource(AsyncResource):
    """Asynchronous Tagger Resource Class."""

    async def tag(
        self,
        text: str,
        task: Optional[Union[Literal["ner"], Literal["acronym-detection"]]] = None,
        model: Optional[str] = None,
        priority: int = 0,
        read_timeout: float = 10.0,
        timeout: float = 180.0,
    ) -> List[TaggerMeta]:
        """Tag a text."""
        data = {
            "input_data": {"docs": [text]},
            "priority": priority,
        }
        if model:
            data["model"] = model
        elif task:
            data["task"] = task
        else:
            raise ValueError("Either `task` or `model` must be provided.")
        output = await self._post(
            data=data,
            read_timeout=read_timeout,
            timeout=timeout,
        )
        output.raise_for_status()
        return [
            TaggerMeta(
                end=entity["char_end_index"],
                label=entity["label"],
                score=entity["score"],
                span=entity["span"],
                start=entity["char_start_index"],
            )
            for entity in output.json()["output"][0]
        ]



"""Ranker Resource Module."""

from typing import Dict, List, Optional

from ..dataclasses.embed import EmbeddingMeta
from .base import AsyncResource, SyncResource


class SyncEmbeddingResource(SyncResource):
    """Synchronous Embedding Resource Class."""

    def embed(
        self,
        texts: List[str],
        model: Optional[str] = None,
        priority: int = 0,
    ) -> List[EmbeddingMeta]:
        """Embed all texts."""
        data: Dict = {
            "input_data": {"texts": texts},
        }
        if model:
            data["model"] = model
        else:
            data["task"] = "embedding"
        output = self._post(
            data={
                "input_data": {"texts": texts},
                "task": "embedding",
                "priority": priority,
            },
        )
        output.raise_for_status()
        return [
            EmbeddingMeta(
                embedding=emb["embedding"],
                text=emb["text"],
            )
            for emb in output.json()["output"]
        ]


class AsyncEmbeddingResource(AsyncResource):
    """Asynchronous Embedding Resource Class."""

    async def embed(
        self,
        texts: List[str],
        model: Optional[str] = None,
        read_timeout: float = 10.0,
        timeout: float = 180.0,
        priority: int = 0,
    ) -> List[EmbeddingMeta]:
        """Asynchronously embed all texts."""
        data: Dict = {
            "input_data": {"texts": texts},
            "priority": priority,
        }
        if model:
            data["model"] = model
        else:
            data["task"] = "embedding"
        response = await self._post(
            data=data,
            read_timeout=read_timeout,
            timeout=timeout,
        )
        response.raise_for_status()  # Ensure proper exception handling in your async context.
        return [
            EmbeddingMeta(
                embedding=emb["embedding"],
                text=emb["text"],
            )
            for emb in response.json()["output"]
        ]



"""Dataclasses for embeddings."""

from typing import List

from pydantic import BaseModel


class EmbeddingMeta(BaseModel):
    """Metadata for an embedding output."""

    text: str
    embedding: List[float]



"""Quasar client: inference, training, etc."""

from typing import List
from urllib.parse import urljoin

import requests
from openai import AsyncOpenAI, OpenAI

from .dataclasses.models import ModelData
from .resources.classifier import AsyncClassifierResource, SyncClassifierResource
from .resources.column_count_classifier import (
    AsyncPageColumnCounterResource,
    SyncPageColumnCounterResource,
)
from .resources.coref import AsyncCorefResource, SyncCorefResource
from .resources.embed import AsyncEmbeddingResource, SyncEmbeddingResource
from .resources.extract import AsyncExtractorResource, SyncExtractorResource

from .resources.multimodal_embed_and_classify import (
    AsyncMultiModalClassifierResource,
    SyncMultiModalClassifierResource,
)
from .resources.rank import AsyncRankerResource, SyncRankerResource
from .resources.tagger import AsyncTaggerResource, SyncTaggerResource


class Quasar(OpenAI):
    """Quasar Client"""

    classifier: SyncClassifierResource
    coref: SyncCorefResource
    embed: SyncEmbeddingResource
    extractor: SyncExtractorResource
    multimodal_embed_and_classify: SyncMultiModalClassifierResource
    page_column_counter: SyncPageColumnCounterResource
    ranker: SyncRankerResource
    tagger: SyncTaggerResource
    
    def __init__(self, quasar_base: str, *args, **kwargs) -> None:
        """Construct an inference client to Quasar."""
        self.quasar_base = quasar_base
        self._models: List[ModelData] = []
        kwargs["api_key"] = "EMPTY"
        # Set OpenAI base_url for OpenAI-compatible methods
        kwargs["base_url"] = urljoin(self.quasar_base, "/predictions/v1")
        super().__init__(*args, **kwargs)
        self.classifier = SyncClassifierResource(self.quasar_base)
        self.coref = SyncCorefResource(self.quasar_base)
        self.embed = SyncEmbeddingResource(self.quasar_base)
        self.extractor = SyncExtractorResource(self.quasar_base)
        self.multimodal_embed_and_classify = SyncMultiModalClassifierResource(self.quasar_base)
        self.page_column_counter = SyncPageColumnCounterResource(self.quasar_base)
        self.ranker = SyncRankerResource(self.quasar_base)
        self.tagger = SyncTaggerResource(self.quasar_base)
        

    def list_models(self, use_cache: bool = False) -> List[ModelData]:
        """Get all models callable within Quasar and cache them."""
        # Early exit to use cache
        if use_cache and self._models:
            return self._models
        list_models_endpoint = urljoin(self.quasar_base, "predictions/v1/models")
        response = requests.get(list_models_endpoint)
        response.raise_for_status()
        models_data = response.json()
        all_models_info = []
        for provider, models in models_data.items():
            for model in models:
                model_id = model.pop("id", None) or model.pop("model", None)
                metadata = model.pop("metadata", {})
                all_models_info.append(
                    ModelData(
                        id=model_id,
                        provider=provider,
                        metadata=metadata,
                    )
                )
        self._models = all_models_info
        return all_models_info

    def list_model_info(self, model_id: str) -> ModelData:
        """List model info."""
        for model in self.list_models(use_cache=True):
            if model.id == model_id:
                return model
        raise ValueError(f"Quasar Model Information {model_id} not found.")


class AsyncQuasar(AsyncOpenAI):
    """Quasar Async Client"""

    classifier: AsyncClassifierResource
    coref: AsyncCorefResource
    embed: AsyncEmbeddingResource
    extractor: AsyncExtractorResource
    multimodal_embed_and_classify: AsyncMultiModalClassifierResource
    page_column_counter: AsyncPageColumnCounterResource
    ranker: AsyncRankerResource
    tagger: AsyncTaggerResource
    
    def __init__(self, quasar_base: str, *args, **kwargs) -> None:
        """Construct an inference client to Quasar."""
        self.quasar_base = quasar_base
        self._models: List[ModelData] = []
        kwargs["api_key"] = "EMPTY"
        # Set OpenAI base_url for OpenAI-compatible methods
        kwargs["base_url"] = urljoin(self.quasar_base, "/predictions/v1")
        super().__init__(*args, **kwargs)
        self.classifier = AsyncClassifierResource(self.quasar_base)
        self.coref = AsyncCorefResource(self.quasar_base)
        self.embed = AsyncEmbeddingResource(self.quasar_base)
        self.extractor = AsyncExtractorResource(self.quasar_base)
        self.multimodal_embed_and_classify = AsyncMultiModalClassifierResource(self.quasar_base)
        self.page_column_counter = AsyncPageColumnCounterResource(self.quasar_base)
        self.ranker = AsyncRankerResource(self.quasar_base)
        self.tagger = AsyncTaggerResource(self.quasar_base)
        
    def list_models(self, use_cache: bool = False) -> List[ModelData]:
        """Get all models callable within Quasar and cache them."""
        # Early exit to use cache
        if use_cache and self._models:
            return self._models
        list_models_endpoint = urljoin(self.quasar_base, "predictions/v1/models")
        response = requests.get(list_models_endpoint)
        response.raise_for_status()
        models_data = response.json()
        all_models_info = []
        for provider, models in models_data.items():
            for model in models:
                model_id = model.pop("id", None) or model.pop("model", None)
                metadata = model.pop("metadata", {})
                all_models_info.append(
                    ModelData(
                        id=model_id,
                        provider=provider,
                        metadata=metadata,
                    )
                )
        self._models = all_models_info
        return all_models_info

    def list_model_info(self, model_id: str) -> ModelData:
        """List model info."""
        for model in self.list_models(use_cache=True):
            if model.id == model_id:
                return model
        raise ValueError(f"Quasar Model Information {model_id} not found.")



"""Dataclasses for coref."""
from typing import List, Optional

from pydantic import BaseModel


class CorefCluster(BaseModel):
    """A coref cluster."""

    start: int
    end: int


class CorefMeta(BaseModel):
    """Metadata for coref output."""

    text: str
    resolved_text: Optional[str] = None
    clusters: Optional[List[List[CorefCluster]]] = None



from abc import ABC, abstractmethod
from typing import Dict
from urllib.parse import urljoin

import httpx
import requests


# Abstract base class
class ResourceBase(ABC):
    """Base class for Quasar resources."""

    def __init__(self, base_url: str):
        self.base_url = base_url
        self.prediction_endpoint = urljoin(self.base_url, "/predictions/")

    @abstractmethod
    def _post(self, data: Dict):
        pass


class SyncResource(ResourceBase):
    """Synchronous resource."""

    def _post(self, data: Dict):
        return requests.post(self.prediction_endpoint, json=data)


class AsyncResource(ResourceBase):
    """Asynchronous resource."""

    async def _post(self, data: Dict, timeout: float = 180.0, read_timeout: float = 10.0):
        timeout = httpx.Timeout(timeout=timeout, read=read_timeout)
        async with httpx.AsyncClient() as client:
            return await client.post(self.prediction_endpoint, json=data, timeout=timeout)



"""Ranker Resource Module."""

from typing import List, Optional, Tuple

from ..dataclasses.rank import RankerMeta
from .base import AsyncResource, SyncResource


class SyncRankerResource(SyncResource):
    """Synchronous Ranker Resource Class."""

    def rank(
        self,
        pairs: List[Tuple[str, str]],
        priority: int = 0,
        model: Optional[str] = None,
    ) -> List[RankerMeta]:
        """Rank pairs of text."""
        data = {
            "input_data": {"pairs": pairs},
            "priority": priority,
        }
        if model:
            data["model"] = model
        else:
            data["task"] = "ranking"
        output = self._post(data=data)
        output.raise_for_status()
        return [
            RankerMeta(
                pair=(pair["query"], pair["candidate"]),
                score=pair["score"],
            )
            for pair in output.json()["output"]
        ]


class AsyncRankerResource(AsyncResource):
    """Asynchronous Ranker Resource Class."""

    async def rank(
        self,
        pairs: List[Tuple[str, str]],
        model: Optional[str] = None,
        read_timeout: float = 10.0,
        timeout: float = 180.0,
        priority: int = 0,
    ) -> List[RankerMeta]:
        """Asynchronously rank pairs of text."""
        data = {
            "input_data": {"pairs": pairs},
            "priority": priority,
        }
        if model:
            data["model"] = model
        else:
            data["task"] = "ranking"
        response = await self._post(
            data=data,
            read_timeout=read_timeout,
            timeout=timeout,
        )
        response.raise_for_status()
        return [
            RankerMeta(
                pair=(pair["query"], pair["candidate"]),
                score=pair["score"],
            )
            for pair in response.json()["output"]
        ]



"""Dataclasses for taggers."""

from pydantic import BaseModel


class TaggerMeta(BaseModel):
    """Metadata for a tagger output."""

    end: int
    label: str
    score: float
    span: str
    start: int



"""Coreference Resolution Resource Module."""

from typing import Optional

from ..dataclasses.coref import CorefCluster, CorefMeta
from .base import AsyncResource, SyncResource


class SyncCorefResource(SyncResource):
    """Synchronous Coref Resource Class."""

    def resolve(
        self,
        text: str,
        model: Optional[str] = None,
        priority: int = 0,
    ) -> CorefMeta:
        """Get corefs and resolved text."""
        data = {
            "input_data": {
                "text": text,
                "resolve_text": True,
            },
            "priority": priority,
        }
        if model:
            data["model"] = model
        else:
            data["task"] = "coreference-resolution"
        output = self._post(data=data)
        output.raise_for_status()
        coref_resp = output.json()["output"]
        clusters = [
            [
                CorefCluster(
                    start=ref[0],
                    end=ref[1],
                )
                for ref in cluster
            ]
            for cluster in coref_resp["coref_clusters"]
        ]
        return CorefMeta(
            text=coref_resp["text"],
            resolved_text=coref_resp["resolved_text"],
            clusters=clusters,
        )


class AsyncCorefResource(AsyncResource):
    """Asynchronous Coref Resource Class."""

    async def resolve(
        self,
        text: str,
        model: Optional[str] = None,
        read_timeout: float = 10.0,
        timeout: float = 180.0,
        priority: int = 0,
    ) -> CorefMeta:
        """Embed all texts."""
        data = {
            "input_data": {"text": text, "resolve_text": True},
            "priority": priority,
        }
        if model:
            data["model"] = model
        else:
            data["task"] = "coreference-resolution"
        output = await self._post(
            data=data,
            read_timeout=read_timeout,
            timeout=timeout,
        )
        output.raise_for_status()
        coref_resp = output.json()["output"]
        clusters = [
            [
                CorefCluster(
                    start=ref[0],
                    end=ref[1],
                )
                for ref in cluster
            ]
            for cluster in coref_resp["coref_clusters"]
        ]
        return CorefMeta(
            text=coref_resp["text"],
            resolved_text=coref_resp["resolved_text"],
            clusters=clusters,
        )



"""Dataclasses for classifiers."""

from typing import List

from pydantic import BaseModel


class Classification(BaseModel):
    """A classification output."""

    label: str
    confidence: float


class ClassifierMeta(BaseModel):
    """Metadata for an classifier output."""

    text: str
    classifications: List[Classification]



"""Dataclasses for extractions."""

from typing import List, Optional

from pydantic import BaseModel


class Keyword(BaseModel):
    """Keyword dataclass."""

    keyword: str
    score: float


class ExtractMeta(BaseModel):
    """Metadata for an extractor output."""

    text: str
    keywords: Optional[List[Keyword]] = None
    topics: Optional[List[str]] = None
